﻿using System.Runtime.InteropServices;
using System.Text;
using System.Xml.Linq;

namespace ArmaRAMDb
{
    public class CallContext
    {
        public ulong SteamId { get; set; }
        public string FileSource { get; set; }
        public string MissionName { get; set; }
        public string ServerName { get; set; }
        public short RemoteExecutedOwner { get; set; }
    }

    public class Main
    {
        private const string ARDB_VERSION = "1.0.0";
        public static string ARDB_LOGFOLDER { get; private set; } = "\\@ramdb\\logs";
        public static bool ARDB_DEBUG {get; private set; } = false;
        public static bool ARDB_INITCHECK {get; private set; } = false;
        public static string STEAMID {get; private set; } = "";
        public static bool ARDB_CONTEXTLOG {get; private set; } = false;
        public static IntPtr ARDB_CONTEXT { get; private set; }
        private static int numCalls = 0;
        public static bool ARDB_ISLOADED { get; private set; } = false;
        public static ExtensionCallback Callback { get; private set; }
        public delegate int ExtensionCallback([MarshalAs(UnmanagedType.LPStr)] string name, [MarshalAs(UnmanagedType.LPStr)] string function, [MarshalAs(UnmanagedType.LPStr)] string data);

        public static void ARDB_Init()
        {
            char[] separator = ['='];
            string[] commandLineArgs = Environment.GetCommandLineArgs();
            string str = "";

            for (int index = 0; index < commandLineArgs.Length; index++)
            {
                string[] strArray = commandLineArgs[index].Split(separator, StringSplitOptions.RemoveEmptyEntries);
                if (strArray[0].Equals("-ardb", StringComparison.CurrentCultureIgnoreCase))
                {
                    str = strArray[1];
                    break;
                }
            }

            if (str == "")
                str = "@ramdb\\config.xml";

            if (File.Exists(Environment.CurrentDirectory + "\\" + str))
            {
                List<string> strList = [];
                List<string> list = XElement.Load(Environment.CurrentDirectory + "\\" + str).Elements().Select(eintrag => (string)eintrag).ToList();
                if (bool.TryParse(list[0], out bool res0))
                    ARDB_CONTEXTLOG = res0;
                if (bool.TryParse(list[1], out bool res1))
                    ARDB_DEBUG = res1;

                Log($"Config file found! Context Mode: {ARDB_CONTEXTLOG}! Debug Mode: {ARDB_DEBUG}!", "action");
            }
            else
            {
                Log("Config file not found! Default Settings Loaded.", "action");
            }

            ARDB_INITCHECK = true;
        }

        public static void Log(string msg, string type)
        {
            if (!ARDB_DEBUG)
                return;
            string logFileName = type + ".log";
            string path = Environment.CurrentDirectory + ARDB_LOGFOLDER;

            if (!Directory.Exists(path))
            {
                Directory.CreateDirectory(path);
            }

            using StreamWriter sw = new(Path.Combine(path, logFileName), true);
            sw.WriteLine(DateTime.Now.ToString() + " >> " + msg);
        }

        /// <summary>
        /// Called only once when Arma 3 loads the extension.
        /// </summary>
        /// <param name="func">Pointer to Arma 3's callback function</param>
        [UnmanagedCallersOnly(EntryPoint = "RVExtensionRegisterCallback")]
        public unsafe static void RVExtensionRegisterCallback(nint func)
        {
            Callback = Marshal.GetDelegateForFunctionPointer<ExtensionCallback>(func);
        }

        /// <summary>
        /// Called only once when Arma 3 loads the extension.
        /// The output will be written in the RPT logs.
        /// </summary>
        /// <param name="output">A pointer to the output buffer</param>
        /// <param name="outputSize">The maximum length of the buffer (always 32 for this particular method)</param>
        [UnmanagedCallersOnly(EntryPoint = "RVExtensionVersion")]
    #pragma warning disable IDE0060 // Remove unused parameter
        public unsafe static void RVExtensionVersion(char* output, int outputSize)
    #pragma warning restore IDE0060 // Remove unused parameter
        {
            WriteOutput(output, ARDB_VERSION);
        }

        /// <summary>
        /// Called before every RVExtension/RVExtensionArgs execution, providing context about where the call came from.
        /// </summary>
        /// <param name="argv">
        /// Array of pointers to null-terminated strings containing:
        /// [0] - Steam 64bit UserID
        /// [1] - FileSource
        /// [2] - MissionName
        /// [3] - ServerName
        /// [4] - Machine Network ID
        /// </param>
        /// <param name="argc">Number of arguments in the argv array (always 5).</param>
        [UnmanagedCallersOnly(EntryPoint = "RVExtensionContext")]
        public unsafe static int RVExtensionContext(char** argv, int argc)
        {
            ARDB_CONTEXT = (IntPtr)argv;

            var context = new CallContext
            {
                SteamId = ulong.Parse(Marshal.PtrToStringAnsi((IntPtr)argv[0])),
                FileSource = Marshal.PtrToStringAnsi((IntPtr)argv[1]),
                MissionName = Marshal.PtrToStringAnsi((IntPtr)argv[2]),
                ServerName = Marshal.PtrToStringAnsi((IntPtr)argv[3]),
                RemoteExecutedOwner = short.Parse(Marshal.PtrToStringAnsi((IntPtr)argv[4]))
            };

            STEAMID = context.SteamId.ToString();

            if (ARDB_CONTEXTLOG)
            {
                Log($"Context: SteamID={context.SteamId}, FileSource={context.FileSource}, MissionName={context.MissionName}, ServerName={context.ServerName}, RemoteExecutedOwner={context.RemoteExecutedOwner}", "context");
            }

            return 100;
        }

        /// <summary>
        /// The entry point for the default "callExtension" command.
        /// </summary>
        /// <param name="output">A pointer to the output buffer</param>
        /// <param name="outputSize">The maximum size of the buffer (20480 bytes)</param>
        /// <param name="function">The function identifier passed in "callExtension"</param>
        [UnmanagedCallersOnly(EntryPoint = "RVExtension")]
    #pragma warning disable IDE0060 // Remove unused parameter
        public unsafe static void RVExtension(char* output, int outputSize, char* function)
    #pragma warning restore IDE0060 // Remove unused parameter
        {
            if (!ARDB_INITCHECK)
                ARDB_Init();

            numCalls++;

            var func = GetString(function);
            switch (func.ToLower())
            {
                case "time":
                    DateTime timeNow = DateTime.Now;
                    string timeNowString = timeNow.ToString("dd:MM:yyyy HH:mm:ss");
                    WriteOutput(output, timeNowString);
                    break;
                default:
                    WriteOutput(output, func);
                    break;
            }
        }

        /// <summary>
        /// The entry point for the "callExtension" command with function arguments.
        /// </summary>
        /// <param name="output">A pointer to the output buffer</param>
        /// <param name="outputSize">The maximum size of the buffer (20480 bytes)</param>
        /// <param name="function">The function identifier passed in "callExtension"</param>
        /// <param name="argv">The args passed to "callExtension" as a string array</param>
        /// <param name="argc">Number of elements in "argv"</param>
        /// <returns>The return code</returns>
        [UnmanagedCallersOnly(EntryPoint = "RVExtensionArgs")]
        public unsafe static int RVExtensionArgs(char* output, int outputSize, char* function, char** argv, int argc)
        {
            if (!ARDB_INITCHECK)
                ARDB_Init();

            numCalls++;

            var _id = DateTimeOffset.Now.ToUnixTimeMilliseconds();
            var func = GetString(function);
            
            List<string> args = Enumerable.Range(0, argc).Select(i => GetString(argv[i])).ToList();

            switch (func.ToLower())
            {
                case "del":
                    HandleDeleteOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "get":
                    HandleGetOperation(args, argc);
                    WriteOutput(output, $"[\"{_id}_get\",{args[0]}]");
                    return 100;
                case "hdel":
                    HandleHashDeleteOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "hdelid":
                    HandleHashDeleteIdOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "hget":
                    HandleHashGetOperation(args, argc);
                    WriteOutput(output, $"[\"{_id}_hget\",{args[1]}]");
                    return 100;
                case "hgetid":
                    HandleHashGetIdOperation(args, argc);
                    WriteOutput(output, $"[\"{_id}_hgetid\",{args[0]}]");
                    return 100;
                case "hgetall":
                    HandleHashGetAllOperation(args, argc);
                    WriteOutput(output, $"[\"{_id}_hgetall\",{args[0]}]");
                    return 100;
                case "hgetallid":
                    HandleHashGetAllIdOperation(args, argc);
                    WriteOutput(output, $"[\"{_id}_hgetallid\",{args[0]}]");
                    return 100;
                case "hrem":
                    HandleHashRemoveOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "hremid":
                    HandleHashRemoveIdOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "hset":
                    HandleHashSetOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "hsetid":
                    HandleHashSetIdOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "hsetbulk":
                    HandleHashSetBulkOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "hsetbulkid":
                    HandleHashSetBulkIdOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "isloaded":
                    WriteOutput(output, ARDB_ISLOADED.ToString().ToLower());
                    return 100;
                case "listadd":
                    HandleListAddOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "listdel":
                    HandleListDeleteOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "listidx":
                    HandleListIndexOperation(args, argc);
                    WriteOutput(output, $"[\"{_id}_listindex\",{args[0]}]");
                    return 100;
                case "listrng":
                    HandleListRangeOperation(args, argc);
                    WriteOutput(output, $"[\"{_id}_listrange\",{args[0]}]");
                    return 100;
                case "listrem":
                    HandleListRemoveOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "listset":
                    HandleListSetOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "load":
                    LoadData();
                    WriteOutput(output, "Data loaded");
                    return 100;
                case "save":
                    SaveData();
                    WriteOutput(output, "Data saved");
                    return 100;
                case "set":
                    HandleSetOperation(args, argc);
                    WriteOutput(output, "Async");
                    return 200;
                case "version":
                    WriteOutput(output, ARDB_VERSION);
                    return 100;
                default:
                    WriteOutput(output, "Available functions: delete, get, hget, hgetid, hgetall, hgetallid, hset, hsetid, hsetbulk, hsetbulkid, isloaded, load, save, set, version");
                    return -1;
            }
        }

        /// <summary>
        /// Reads a string from the given pointer. 
        /// If the pointer is null, a default value will be returned instead.
        /// </summary>
        /// <param name="pointer">The pointer to read</param>
        /// <param name="defaultValue">The string's default value</param>
        /// <returns>The marshalled string</returns>
        private unsafe static string GetString(char* pointer, string defaultValue = "")
        {
            return Marshal.PtrToStringAnsi((nint)pointer) ?? defaultValue;
        }

        /// <summary>
        /// Serializes a list of strings into a string representing a valid Arma 3 array.
        /// </summary>
        /// <param name="list">The list of strings to serialize</param>
        /// <returns>A string representing an Arma 3 array</returns>
        private static string SerializeList(List<string> list)
        {
            var content = string.Join(",", [.. list]);
            return string.Format("[{0}]", content);
        }

        /// <summary>
        /// Writes the given payload to the output buffer using the provided pointer.
        /// Ensures that the payload string is always null terminated.
        /// </summary>
        /// <param name="output">A pointer to the output buffer</param>
        /// <param name="payload">The payload to write</param>
        private unsafe static void WriteOutput(char* output, string payload)
        {
            byte[] bytes = Encoding.ASCII.GetBytes(payload + '\0');
            Marshal.Copy(bytes, 0, (nint)output, bytes.Length);
        }

        private static void SaveData()
        {
            var db = new RAMDb();

            db.ExportToXml();
            Log("Data saved to XML", "action");
        }

        private static void LoadData()
        {
            var db = new RAMDb();

            db.ImportFromXml();
            ARDB_ISLOADED = true;
            Log("Data loaded from XML", "action");
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleHashDeleteOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            HashStore.HashDelete(STEAMID);
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleHashDeleteIdOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            HashStore.HashDelete(args[0]);
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleHashGetOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var value = HashStore.HashGet(STEAMID, args[0]);

            Log($"HashGet: {SerializeList([value])}", "debug");
            Callback("ArmaRAMDb", args[1].Trim('"'), SerializeList([value]));
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleHashGetIdOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var value = HashStore.HashGet(args[0], args[1]);

            Log($"HashGetId: {SerializeList([value])}", "debug");
            Callback("ArmaRAMDb", args[2].Trim('"'), SerializeList([value]));
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleHashGetAllOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var contents = HashStore.HashGetAll(STEAMID);
            var list = contents.SelectMany(kv => new[] { kv.Key, kv.Value }).ToList();

            Log($"HashGetAll: {SerializeList(list)}", "debug");
            Callback("ArmaRAMDb", args[0].Trim('"'), SerializeList(list));
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleHashGetAllIdOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var hashId = args[0];
            var contents = HashStore.HashGetAll(hashId);
            var list = contents.SelectMany(kv => new[] { kv.Key, kv.Value }).ToList();

            Log($"HashGetAllId: {SerializeList(list)}", "debug");
            Callback("ArmaRAMDb", args[1].Trim('"'), SerializeList(list));
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleHashRemoveOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            HashStore.HashRemove(STEAMID, args[0]);
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleHashRemoveIdOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var value = HashStore.HashRemove(args[0], args[1]);
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleHashSetOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            Log($"HashSet: {SerializeList(args)}", "debug");
            HashStore.HashSet(STEAMID, args[0], args[1]);
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleHashSetIdOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            Log($"HashSetId: {SerializeList(args)}", "debug");
            HashStore.HashSet(args[0], args[1], args[2]);
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleHashSetBulkOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var values = new Dictionary<string, string>();
    
            for (int i = 0; i < args.Count; i += 2)
            {
                string key = args[i];
                string value = args[i + 1];
                values.Add(key, value);
            }
            
            Log($"BulkHashSet: Keys={string.Join(",", values.Keys)}", "debug");
            HashStore.HashSetBulk(STEAMID, values);
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleHashSetBulkIdOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var hashId = args[0];
            var values = new Dictionary<string, string>();
    
            for (int i = 1; i < args.Count; i += 2)
            {
                string key = args[i];
                string value = args[i + 1];
                values.Add(key, value);
            }
            
            Log($"BulkHashSetId: Keys={string.Join(",", values.Keys)}", "debug");
            HashStore.HashSetBulk(hashId, values);
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleListAddOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var key = args[0];
            var value = args[1];

            Log($"ListAdd: {SerializeList([key, value])}", "debug");
            ListStore.ListAdd(key, value);
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleListDeleteOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var listId = args[0];

            Log($"ListDelete: {SerializeList([listId])}", "debug");
            ListStore.ListDelete(listId);
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleListIndexOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var listId = args[0];
            var key = args[1];
            var value = ListStore.ListIndex(listId, int.Parse(key));
            
            Log($"ListIndex: {SerializeList([key, value])}", "debug");
            Callback("ArmaRAMDb", args[2].Trim('"'), SerializeList([value]));
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleListRangeOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var listId = args[0];
            var startIndex = int.Parse(args[1]);
            var endIndex = int.Parse(args[2]);
            
            var values = ListStore.ListRange(listId, startIndex, endIndex);
            
            Log($"ListRange: ListId={listId} Start={startIndex} End={endIndex} Values={SerializeList(values)}", "debug");
            Callback("ArmaRAMDb", args[3].Trim('"'), SerializeList(values));
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleListRemoveOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var listId = args[0];
            var index = int.Parse(args[1]);

            ListStore.ListSet(listId, index, "RAMDBREMOVE");
            ListStore.ListRemove(listId, "RAMDBREMOVE");
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleListSetOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var listId = args[0];
            var index = int.Parse(args[1]);
            var value = args[2];

            ListStore.ListSet(listId, index, value);
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleSetOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var key = args[0];
            var value = args[1];
            
            KeyValueStore.SetValue(key, value);
            Log($"Set: Key={key} Value={value}", "debug");
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleGetOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var key = args[0];
            var value = KeyValueStore.GetValue(key);
            
            Log($"Get: Key={key} Value={value}", "debug");
            Callback("ArmaRAMDb", args[1].Trim('"'), SerializeList([value]));
        }

#pragma warning disable IDE0060 // Remove unused parameter
        private static void HandleDeleteOperation(List<string> args, int argc)
#pragma warning restore IDE0060 // Remove unused parameter
        {
            var key = args[0];
            
            KeyValueStore.DeleteValue(key);
            Log($"Delete: Key={key}", "debug");
        }
    }
}